\name{bhlm}
\alias{bhlm}

\title{Bayesian Hierarchical Linear Mixed Effects Model}

\description{
Fit a Bayesian hierarchical linear mixed effects model of the form:

  Y_ij = X_ij beta_j + M_ij gamma + e_ij

  beta_j = Z_j alpha + u_j   \bold{:the random effects.}

  j = 1, ..., J: the group index
    
  i = 1, ..., n_j: the observation index within group j
    
  e_ij ~ p( e_ij; sigma^2_j )   \bold{:the error distribution.  May be normal or t.}

  u_jk ~ N( 0, tau^2_k )  \bold{OR:}  u_j ~ N( 0, V )  

  gamma ~ p( gamma )   \bold{:the fixed effects.}

  alpha ~ p( alpha )   \bold{:the second-level effects.}

  where the tau^2_k are given univariate priors or V is given an inverse Wishart prior.

  See \code{\link{bhlm.prior}} for details.
  
  The distribution p( gamma ) can be equal to N( gamma_0, V1_0 ) or t( df_gamma, gamma_0, V1_0 ), or it can be flat (improper).  The same holds for p( alpha ).
}

\usage{
bhlm(fixed.formula = NULL, random.formula = NULL, level2.formula = NULL,
     group.formula = NULL, data, prior = bhlm.prior(),
     likelihood  = bhlm.likelihood(), sampler = bhlm.sampler(),
     na.action = NULL, contrasts = NULL)
}

\arguments{
  \item{fixed.formula}{an object of class \code{formula} specifying the fixed effects predictors and possibly the outcome, e.g. \code{random.formula = Y ~ M1}, indicates that the intercept and \code{M1} are considered fixed effects predictors and that \code{Y} is the outcome variable.  If the outcome variable is not specified in \code{random.formula} then it must be specified in \code{fixed.formula}.}

  \item{random.formula}{an object of class \code{formula} specifying the random effects predictors and possibly the outcome variable, e.g. \code{random.formula = Y ~ X1 + X2 - 1} indicates that the random effects predictors are \code{X1} and \code{X2}, that the intercept is not to be considered as a random effect, and that \code{Y} is the outcome variable.}

  \item{level2.formula}{an object of class \code{formula} specifying the second level effects predictors. These predictors are assumed constant within groups, e.g. \code{level2.formula  = ~ 1}, specifies a common population mean model.}

  \item{group.formula}{an object of class \code{formula} specifying the variable indicating the groups, e.g. \code{group.formula = ~ School}, indicates that all those cases (observations) within the same \code{School} values constitute the groups in the hierarchical model.  If \code{group.formula = NULL}, all of the observations are taken to be in the same group.}

  \item{data}{an object of class \code{data.frame} specifying the data.}

  \item{prior}{a list specifying the prior distributions, generated by a call to \code{\link{bhlm.prior}}.}

  \item{likelihood}{a list specifying the likelihood structure of a Bayesian hierarchical linear model, generated by a call to \code{\link{bhlm.likelihood}}.}

  \item{sampler}{a list specifying the number of MCMC chains as well as operating characteristics and initial values for the chains, generated by a call to \code{\link{bhlm.sampler}}.}

  \item{na.action}{a flag that tells what to do with missing data.  The default \code{na.action = na.fail} causes the program to fail on encountering missing data.  Another valid value of \code{na.action} is \code{na.omit}.}

  \item{contrasts}{argument to pass to a \code{model.matrix} call.}
}

\details{
The posterior distribution of the parameters of the model are estimated via Markov Chain Monte Carlo simulation (a Gibbs sampler is performed).
}

\section{References}{
Gelman, A., J. B. Carlin, H. S. Stern, and D. B. Rubin (2004). \emph{Bayesian Data Analysis}, 2nd Edition. Boca Raton, Florida: Chapman & Hall.

Morris, C. N., and S. L. Normand, "Hierarchical models for combining information and for meta-analyses", In \emph{Bayesian Statistics 4, J. M. Bernardo, J. O. Berger, A. P. Dawid, and F. M. Smith Eds.},  pages 321-344, Oxford University Press, 1992.

Hobert, J. P., and G. Casella, "The effect of improper priors on Gibbs sampling in hierarchical linear mixed models", \emph{J. American Statistical Association}, vol. 91, 436: 1461- 1473, 1996.

 Daniels, M. J., "A prior for the variance in hierarchical models", \emph{The Canadian Journal of Statistics}, vol. 27, 3: 567-578, 1999.

Wakefield, J. C., F. M. Smith, A. Racine-Poon, and A. E. Gelfand, "Bayesian analysis of linear and non-linear population models by using the Gibbs sampler", \emph{Journal of the Royal Statistical Society Series C (Appl. Statist.)}, vol. 43, 1: 201-221, 1994.
}

\seealso{
\code{\link{bhlm.prior}},
\code{\link{bhlm.likelihood}},
\code{\link{bhlm.sampler}}.
}

\keyword{FlexBayes}
\keyword{prior}


